'use strict';

module.exports = {
	acosh:  require('./acosh'),
	asinh:  require('./asinh'),
	atanh:  require('./atanh'),
	cbrt:   require('./cbrt'),
	clz32:  require('./clz32'),
	cosh:   require('./cosh'),
	expm1:  require('./expm1'),
	fround: require('./fround'),
	hypot:  require('./hypot'),
	imul:   require('./imul'),
	log10:  require('./log10'),
	log2:   require('./log2'),
	log1p:  require('./log1p'),
	sign:   require('./sign'),
	sinh:   require('./sinh'),
	tanh:   require('./tanh'),
	trunc:  require('./trunc')
};
